@@ -486,38 +486,65 @@ int xc_domain_pin_memory_cacheattr(xc_interface *xch,
 }
 
 #if defined(__i386__) || defined(__x86_64__)
-#include "xc_e820.h"
-int xc_domain_set_memmap_limit(xc_interface *xch,
+int xc_domain_set_memory_map(xc_interface *xch,
                                uint32_t domid,
-                               unsigned long map_limitkb)
+                               struct e820entry entries[],
+                               uint32_t nr_entries)
 {
     int rc;
     struct xen_foreign_memory_map fmap = {
         .domid = domid,
-        .map = { .nr_entries = 1 }
+        .map = { .nr_entries = nr_entries }
     };
-    DECLARE_HYPERCALL_BUFFER(struct e820entry, e820);
+    DECLARE_HYPERCALL_BOUNCE(entries, nr_entries * sizeof(struct e820entry),
+                             XC_HYPERCALL_BUFFER_BOUNCE_IN);
 
-    e820 = xc_hypercall_buffer_alloc(xch, e820, sizeof(*e820));
-
-    if ( e820 == NULL )
-    {
-        PERROR("Could not allocate memory for xc_domain_set_memmap_limit hypercall");
+    if ( !entries || xc_hypercall_bounce_pre(xch, entries) )
         return -1;
-    }
 
-    e820->addr = 0;
-    e820->size = (uint64_t)map_limitkb << 10;
-    e820->type = E820_RAM;
-
-    set_xen_guest_handle(fmap.map.buffer, e820);
+    set_xen_guest_handle(fmap.map.buffer, entries);
 
     rc = do_memory_op(xch, XENMEM_set_memory_map, &fmap, sizeof(fmap));
 
-    xc_hypercall_buffer_free(xch, e820);
+    xc_hypercall_bounce_post(xch, entries);
 
     return rc;
 }
+int xc_get_machine_memory_map(xc_interface *xch,
+                              struct e820entry entries[],
+                              uint32_t max_entries)
+{
+    int rc;
+    struct xen_memory_map memmap = {
+        .nr_entries = max_entries
+    };
+    DECLARE_HYPERCALL_BOUNCE(entries, sizeof(struct e820entry) * max_entries,
+                             XC_HYPERCALL_BUFFER_BOUNCE_OUT);
+
+    if ( !entries || xc_hypercall_bounce_pre(xch, entries) || max_entries <= 1)
+        return -1;
+
+
+    set_xen_guest_handle(memmap.buffer, entries);
+
+    rc = do_memory_op(xch, XENMEM_machine_memory_map, &memmap, sizeof(memmap));
+
+    xc_hypercall_bounce_post(xch, entries);
+
+    return rc ? rc : memmap.nr_entries;
+}
+int xc_domain_set_memmap_limit(xc_interface *xch,
+                               uint32_t domid,
+                               unsigned long map_limitkb)
+{
+    struct e820entry e820;
+
+    e820.addr = 0;
+    e820.size = (uint64_t)map_limitkb << 10;
+    e820.type = E820_RAM;
+
+    return xc_domain_set_memory_map(xch, domid, &e820, 1);
+}
 #else
 int xc_domain_set_memmap_limit(xc_interface *xch,
                                uint32_t domid,
@@ -913,6 +940,81 @@ int xc_domain_get_pod_target(xc_interface *xch,
                                 pod_entries);
 }
 
+int xc_domain_memory_translate_gpfn_list(xc_interface *xch,
+                                        uint32_t domid,
+                                        unsigned long nr_gpfns,
+                                        xen_pfn_t *gpfn_list,
+                                        xen_pfn_t *mfn_list)
+{
+    int err = - 1;
+    DECLARE_HYPERCALL_BOUNCE(gpfn_list, nr_gpfns*sizeof(*gpfn_list), XC_HYPERCALL_BUFFER_BOUNCE_IN);
+    DECLARE_HYPERCALL_BOUNCE(mfn_list, nr_gpfns*sizeof(*mfn_list), XC_HYPERCALL_BUFFER_BOUNCE_OUT);
+    struct xen_translate_gpfn_list translate_gpfn_list = {
+        .domid    = domid,
+        .nr_gpfns = nr_gpfns,
+    };
+
+    if ( xc_hypercall_bounce_pre(xch, gpfn_list) ||
+         xc_hypercall_bounce_pre(xch, mfn_list))
+    {
+        PERROR("Could not bounce memory for XENMEM_translate_gpfn_list hypercall");
+        return -1;
+    }
+
+    set_xen_guest_handle(translate_gpfn_list.gpfn_list, gpfn_list);
+    set_xen_guest_handle(translate_gpfn_list.mfn_list, mfn_list);
+
+    err = do_memory_op(xch, XENMEM_translate_gpfn_list, 
+                       &translate_gpfn_list, sizeof(translate_gpfn_list));
+
+    if ( err != 0 )
+    {
+        errno = -err;
+        err = -1;
+    }
+
+    xc_hypercall_bounce_post(xch, gpfn_list);
+    xc_hypercall_bounce_post(xch, mfn_list);
+
+    return err;
+}
+
+int xc_domain_memory_release_mfn_list(xc_interface *xch,
+                                      uint32_t domid,
+                                      unsigned long nr_mfns,
+                                      xen_pfn_t *mfn_list)
+{
+    int err;
+    DECLARE_HYPERCALL_BOUNCE(mfn_list, nr_mfns*sizeof(*mfn_list), XC_HYPERCALL_BUFFER_BOUNCE_IN);
+    struct xen_release_mfn_list release_mfn_list = {
+        .domid    = domid,
+        .nr_mfns = nr_mfns,
+    };
+
+    if (xc_hypercall_bounce_pre(xch, mfn_list))
+    {
+        PERROR("Could not bounce memory for XENMEM_release_mfn_list hypercall");
+        return -1;
+    }
+
+    set_xen_guest_handle(release_mfn_list.mfn_list, mfn_list);
+
+    err = do_memory_op(xch, XENMEM_release_mfn_list, 
+                       &release_mfn_list, sizeof(release_mfn_list));
+
+    if ( err != 0 )
+    {
+        DPRINTF("Failed mfns release for dom %d (%ld MFNs)\n",
+                domid, nr_mfns);
+        errno = -err;
+        err = -1;
+    }
+
+    xc_hypercall_bounce_post(xch, mfn_list);
+
+    return err;
+}
+
 int xc_domain_max_vcpus(xc_interface *xch, uint32_t domid, unsigned int max)
 {
     DECLARE_DOMCTL;
@@ -1113,6 +1215,123 @@ int xc_get_hvm_param(xc_interface *handle, domid_t dom, int param, unsigned long
     return 0;
 }
 
+ioservid_or_error_t xc_hvm_register_ioreq_server(xc_interface *xch,
+                                                 domid_t dom)
+{
+    xen_hvm_register_ioreq_server_t arg;
+    ioservid_or_error_t rc = -1;
+
+    arg.domid = dom;
+
+    rc = do_hvm_op(xch, HVMOP_register_ioreq_server, &arg);
+    if (rc)
+        return rc;
+
+    rc = arg.id;
+    return rc;
+}
+
+unsigned long xc_hvm_get_ioreq_server_buf_pfn(xc_interface *xch,
+                                              domid_t dom,
+                                              ioservid_t id)
+{
+    unsigned long pfn;
+
+    if ( id == 0 )
+    {
+        xc_get_hvm_param(xch, dom, HVM_PARAM_IOREQ0_BUF_PFN, &pfn);
+    }
+    else
+    {
+        xc_get_hvm_param(xch, dom, HVM_PARAM_IOREQN_BUF_PFN, &pfn);
+        pfn += id - 1;
+    }
+
+    return pfn;
+}
+
+unsigned long xc_hvm_get_ioreq_server_shared_pfn(xc_interface *xch,
+                                                 domid_t dom,
+                                                 ioservid_t id)
+{
+    unsigned long pfn;
+
+    if ( id == 0 )
+    {
+        xc_get_hvm_param(xch, dom, HVM_PARAM_IOREQ0_SHARED_PFN, &pfn);
+    }
+    else
+    {
+        xc_get_hvm_param(xch, dom, HVM_PARAM_IOREQN_SHARED_PFN, &pfn);
+        pfn += id - 1;
+    }
+
+    return pfn;
+}
+
+evtchn_port_or_error_t xc_hvm_get_ioreq_server_channel(xc_interface *xch,
+                                                       domid_t dom,
+                                                       ioservid_t id)
+{
+    xen_hvm_get_ioreq_server_channel_t arg;
+    evtchn_port_or_error_t rc = -1;
+
+    arg.domid = dom;
+    arg.id = id;
+
+    rc = do_hvm_op(xch, HVMOP_get_ioreq_server_channel, &arg);
+    if (rc)
+        return rc;
+
+    rc = arg.channel;
+    return rc;
+}
+
+int xc_hvm_map_io_range_to_ioreq_server(xc_interface *xch, domid_t dom,
+                                        ioservid_t id, int is_mmio,
+                                        uint64_t start, uint64_t end)
+{
+    xen_hvm_map_io_range_to_ioreq_server_t  arg;
+
+    arg.domid = dom;
+    arg.id = id;
+    arg.is_mmio = is_mmio;
+    arg.start = start;
+    arg.end = end;
+
+    return do_hvm_op(xch, HVMOP_map_io_range_to_ioreq_server, &arg);
+}
+
+int xc_hvm_unmap_io_range_from_ioreq_server(xc_interface *xch, domid_t dom,
+                                            ioservid_t id, int is_mmio,
+                                            uint64_t start)
+{
+    xen_hvm_unmap_io_range_from_ioreq_server_t  arg;
+
+    arg.domid = dom;
+    arg.id = id;
+    arg.is_mmio = is_mmio;
+    arg.start = start;
+
+    return do_hvm_op(xch, HVMOP_unmap_io_range_from_ioreq_server, &arg);
+}
+
+int xc_hvm_map_pcidev_to_ioreq_server(xc_interface *xch, domid_t dom,
+                                      ioservid_t id, uint16_t domain,
+                                      uint8_t bus, uint8_t device, uint8_t function)
+{
+    xen_hvm_map_pcidev_to_ioreq_server_t arg;
+
+    arg.domid = dom;
+    arg.id = id;
+    arg.domain = domain;
+    arg.bus = bus;
+    arg.device = device;
+    arg.function = function;
+
+    return do_hvm_op(xch, HVMOP_map_pcidev_to_ioreq_server, &arg);
+}
+
 int xc_domain_setdebugging(xc_interface *xch,
                            uint32_t domid,
                            unsigned int enable)
