@@ -32,7 +32,6 @@
 #include "xg_save_restore.h"
 
 #include <xen/hvm/params.h>
-#include "xc_e820.h"
 
 /*
 ** Default values for important tuning parameters. Can override by passing
@@ -905,7 +904,8 @@ static int save_tsc_info(xc_interface *xch, uint32_t dom, int io_fd)
 
 int xc_domain_save(xc_interface *xch, int io_fd, uint32_t dom, uint32_t max_iters,
                    uint32_t max_factor, uint32_t flags,
-                   struct save_callbacks* callbacks, int hvm)
+                   struct save_callbacks* callbacks, int hvm,
+                   uint64_t vm_genid_addr)
 {
     xc_dominfo_t info;
     DECLARE_DOMCTL;
@@ -1355,8 +1355,8 @@ int xc_domain_save(xc_interface *xch, int io_fd, uint32_t dom, uint32_t max_iter
 
                     DDPRINTF(tdom, "map fail: page %i mfn %08lx err %d\n",
                             j, gmfn, pfn_err[j]);
-                    pfn_type[j] = XEN_DOMCTL_PFINFO_XTAB;
-                    continue;
+                    errno = -pfn_err[j];
+                    goto out;
                 }
 
                 if ( pfn_type[j] == XEN_DOMCTL_PFINFO_XTAB )
@@ -1619,6 +1619,19 @@ int xc_domain_save(xc_interface *xch, int io_fd, uint32_t dom, uint32_t max_iter
             uint64_t data;
         } chunk = { 0, };
 
+        if ( vm_genid_addr )
+        {
+            chunk.id = XC_SAVE_ID_HVM_GENERATION_ID_ADDR;
+            chunk.data = vm_genid_addr;
+            if ( wrexact(io_fd, &chunk, sizeof(chunk)) )
+            {
+                DPERROR(tdom, "Error when writing the generation ID address");
+                goto out;
+            }
+        }
+        else
+            DDPRINTF(tdom, "No generation ID address");
+
         chunk.id = XC_SAVE_ID_HVM_IDENT_PT;
         chunk.data = 0;
         xc_get_hvm_param(xch, dom, HVM_PARAM_IDENT_PT,
@@ -1707,9 +1720,9 @@ int xc_domain_save(xc_interface *xch, int io_fd, uint32_t dom, uint32_t max_iter
 
         /* Save magic-page locations. */
         memset(magic_pfns, 0, sizeof(magic_pfns));
-        xc_get_hvm_param(xch, dom, HVM_PARAM_IOREQ_PFN,
+        xc_get_hvm_param(xch, dom, HVM_PARAM_IOREQ0_SHARED_PFN,
                          (unsigned long *)&magic_pfns[0]);
-        xc_get_hvm_param(xch, dom, HVM_PARAM_BUFIOREQ_PFN,
+        xc_get_hvm_param(xch, dom, HVM_PARAM_IOREQ0_BUF_PFN,
                          (unsigned long *)&magic_pfns[1]);
         xc_get_hvm_param(xch, dom, HVM_PARAM_STORE_PFN,
                          (unsigned long *)&magic_pfns[2]);
