@@ -538,6 +538,21 @@ int do_hvm_op(xc_interface *xch, unsigned cmd, void *arg)
     case HVMOP_track_dirty_vram:
         arg_size = sizeof(pht.u.track_dirty_vram);
         break;
+    case HVMOP_register_ioreq_server:
+        arg_size = sizeof(pht.u.register_ioreq_server);
+        break;
+    case HVMOP_get_ioreq_server_channel:
+        arg_size = sizeof(pht.u.get_ioreq_server_channel);
+        break;
+    case HVMOP_map_io_range_to_ioreq_server:
+        arg_size = sizeof(pht.u.map_io_range_to_ioreq_server);
+        break;
+    case HVMOP_unmap_io_range_from_ioreq_server:
+        arg_size = sizeof(pht.u.unmap_io_range_from_ioreq_server);
+        break;
+    case HVMOP_map_pcidev_to_ioreq_server:
+        arg_size = sizeof(pht.u.map_pcidev_to_ioreq_server);
+        break;
     default:
         ERROR("do_hvm_op: WARNING: unknown op=%d", cmd);
         return -1;
@@ -641,8 +656,18 @@ int do_memory_op(xc_interface *xch, int cmd, void *arg, size_t len)
     case XENMEM_set_memory_map:
         arg_size = sizeof(pmt.u.foreign_memory_map);
         break;
+    case XENMEM_translate_gpfn_list:
+        arg_size = sizeof(pmt.u.translate_gpfn_list);
+        break;
+    case XENMEM_release_mfn_list:
+        arg_size = sizeof(pmt.u.release_mfn_list);
+        break;
     default:
-        ERROR("do_memory_op: WARNING: op=%d not converted to safe ioctls", cmd);
+        /*
+         * No IOCTL_PRIVCMD_MEMOP implementation for this sub-op.
+         * Fallback to a direct hypercall.  This will fail with
+         * -EACCES on restricted fds.
+         */
         errno = EINVAL;
         arg_size = len;
         break;
@@ -929,6 +954,8 @@ const char *xc_strerror(xc_interface *xch, int errcode)
         errbuf = pthread_getspecific(errbuf_pkey);
         if (errbuf == NULL) {
             errbuf = malloc(XS_BUFSIZE);
+            if ( errbuf == NULL )
+                return "(failed to allocate errbuf)";
             pthread_setspecific(errbuf_pkey, errbuf);
         }
 
