@@ -32,6 +32,9 @@
 /* Maximum we can support with current vLAPIC ID mapping. */
 #define HVM_MAX_VCPUS        128
 
+/* Maximum number we want to allocate special pages for. */
+#define HVM_MAX_EMULATORS    2
+
 #define HVM_INFO_TABLE_HAS_S3_ENABLED 1
 #define HVM_INFO_TABLE_HAS_S4_ENABLED 1
 
