@@ -49,10 +49,6 @@
 
 #define HVM_PARAM_PAE_ENABLED  4
 
-#define HVM_PARAM_IOREQ_PFN    5
-
-#define HVM_PARAM_BUFIOREQ_PFN 6
-
 #ifdef __ia64__
 
 #define HVM_PARAM_NVRAM_FD     7
@@ -104,9 +100,6 @@
 /* Identity-map page directory used by Intel EPT when CR0.PG=0. */
 #define HVM_PARAM_IDENT_PT     12
 
-/* Device Model domain, defaults to 0. */
-#define HVM_PARAM_DM_DOMAIN    13
-
 /* ACPI S state: currently support S0 and S3 on x86. */
 #define HVM_PARAM_ACPI_S_STATE 14
 
@@ -145,6 +138,13 @@
 #define HVMPME_mode_sync       2
 #define HVMPME_onchangeonly    (1 << 2)
 
-#define HVM_NR_PARAMS          26
+/* Param for ioreq servers */
+#define HVM_PARAM_IOREQ0_BUF_PFN    40
+#define HVM_PARAM_IOREQ0_SHARED_PFN 41
+#define HVM_PARAM_IOREQN_BUF_PFN    42
+#define HVM_PARAM_IOREQN_SHARED_PFN 43
+#define HVM_PARAM_MAX_NR_EMULATORS  44
+
+#define HVM_NR_PARAMS           45
 
 #endif /* __XEN_PUBLIC_HVM_PARAMS_H__ */
