@@ -40,7 +40,12 @@ extern struct domain *dom0;
 #else
 #define BITS_PER_EVTCHN_WORD(d) (has_32bit_shinfo(d) ? 32 : BITS_PER_LONG)
 #endif
-#define MAX_EVTCHNS(d) (BITS_PER_EVTCHN_WORD(d) * BITS_PER_EVTCHN_WORD(d))
+/*
+ * How many words is each bit of vcpu_info->evtchn_pending_sel for?
+ */
+#define WORDS_PER_EVTCHN_SEL_BIT(d) ((d)->domain_id == 0 ? 4 : 1)
+#define MAX_EVTCHNS(d) ( BITS_PER_EVTCHN_WORD(d) * BITS_PER_EVTCHN_WORD(d) \
+                         * WORDS_PER_EVTCHN_SEL_BIT(d) )
 #define EVTCHNS_PER_BUCKET 128
 #define NR_EVTCHN_BUCKETS  (NR_EVENT_CHANNELS / EVTCHNS_PER_BUCKET)
 
@@ -143,6 +148,9 @@ struct vcpu
     bool_t           defer_shutdown;
     /* VCPU is paused following shutdown request (d->is_shutting_down)? */
     bool_t           paused_for_shutdown;
+    /* VCPU need affinity restored */
+    bool_t           affinity_broken;
+
 
     /*
      * > 0: a single port is being polled;
@@ -165,6 +173,8 @@ struct vcpu
     cpumask_t        cpu_affinity;
     /* Used to change affinity temporarily. */
     cpumask_t        cpu_affinity_tmp;
+    /* Used to restore affinity across S3. */
+    cpumask_t        cpu_affinity_saved;
 
     /* Bitmask of CPUs which are holding onto this VCPU's state. */
     cpumask_t        vcpu_dirty_cpumask;
@@ -628,6 +638,7 @@ int schedule_cpu_switch(unsigned int cpu, struct cpupool *c);
 void vcpu_force_reschedule(struct vcpu *v);
 int cpu_disable_scheduler(unsigned int cpu);
 int vcpu_set_affinity(struct vcpu *v, cpumask_t *affinity);
+void restore_vcpu_affinity(struct domain *d);
 
 void vcpu_runstate_get(struct vcpu *v, struct vcpu_runstate_info *runstate);
 void domain_runstate_get(struct domain *d, domain_runstate_info_t *runstate);
